import random

from typing import List, Optional

from bitcoin_coin_selection.selection_types.change_constants import MIN_CHANGE
from bitcoin_coin_selection.selection_types.coin_selection import CoinSelection

from bitcoin_coin_selection.selection_types.output_group import OutputGroup


DEFAULT_ITERATIONS = 1000


def approximate_best_subset(utxo_pool: List[OutputGroup], target_value: int,
                            total_lower: int, iterations: int) -> CoinSelection:

    best_selection = [True for output_group in utxo_pool]
    best_value = total_lower

    for iteration_number in range(iterations):
        if best_value == target_value:
            break
        included = [False for output_group in utxo_pool]
        total_value = 0
        reached_target = False
        number_of_passes = 2
        for pass_number in range(number_of_passes):
            if reached_target:
                break
            for i in range(len(utxo_pool)):
                # The solver here uses a randomized algorithm,
                # the randomness serves no real security purpose but is just
                # needed to prevent degenerate behavior and it is important
                # that the rng is fast. We do not use a constant random sequence,
                # because there may be some privacy improvement by making
                # the selection random.
                if (random.choice([True, False]) if pass_number == 0 else not included[i]):
                    total_value += utxo_pool[i].effective_value
                    included[i] = True
                    if total_value >= target_value:
                        reached_target = True
                        if total_value < best_value:
                            best_value = total_value
                            best_selection = included.copy()
                        total_value -= utxo_pool[i].effective_value
                        included[i] = False

    if reached_target:
        return CoinSelection.from_utxo_pool(best_selection, utxo_pool)
    else:
        return CoinSelection.algorithm_failure()


def select_coins_knapsack_solver(
        utxo_pool: List[OutputGroup], target_value: int,
        iterations=DEFAULT_ITERATIONS) -> CoinSelection:

    # lowest output group larger than target_value
    lowest_larger: Optional[OutputGroup] = None
    applicable_groups: List[OutputGroup] = []
    total_lower = 0

    random.shuffle(utxo_pool)

    for output_group in utxo_pool:
        if output_group.effective_value == target_value:
            return CoinSelection([output_group])

        elif output_group.effective_value < target_value + MIN_CHANGE:
            applicable_groups.append(output_group)
            total_lower += output_group.effective_value

        elif not lowest_larger or output_group.effective_value < lowest_larger.effective_value:
            lowest_larger = output_group

    if total_lower == target_value:
        return CoinSelection(applicable_groups)

    if total_lower < target_value and lowest_larger:
        return CoinSelection([lowest_larger])

    # Solve subset sum by stochastic approximation
    utxo_pool.sort(reverse=True)
    best_selection = approximate_best_subset(
        applicable_groups, target_value, total_lower, iterations
    )
    if best_selection.effective_value != target_value and total_lower >= target_value + MIN_CHANGE:
        best_selection = approximate_best_subset(
            applicable_groups, target_value + MIN_CHANGE, total_lower, iterations
        )
    # If we have a bigger coin and (either the stochastic approximation didn't find
    # a good solution, or the next bigger coin is closer), return the bigger coin
    if lowest_larger and (
        (
            best_selection.effective_value != target_value
            and best_selection.effective_value < target_value + MIN_CHANGE
        )
        or
        (
            lowest_larger.effective_value <= best_selection.effective_value
        )
    ):
        best_selection = CoinSelection([lowest_larger])

    return best_selection



from typing import List

from bitcoin_coin_selection.selection_types.change_constants import MAX_MONEY
from bitcoin_coin_selection.selection_types.output_group import OutputGroup
from bitcoin_coin_selection.selection_types.coin_selection import CoinSelection

TOTAL_TRIES = 100000


def select_coins_branch_and_bound(utxo_pool: List[OutputGroup], target_value: int,
                                  cost_of_change: int) -> CoinSelection:

    current_value = 0
    current_selection: List[bool] = []
    current_available_value = sum(
        [output_group.effective_value for output_group in utxo_pool])
    current_waste = 0

    best_waste = MAX_MONEY
    best_selection: List[bool] = []

    utxo_pool.sort(reverse=True)

    for i in range(TOTAL_TRIES):
        should_backtrack = False

        if (
            # Cannot possibly reach target with the amount remaining in the curr_available_value
            (current_value + current_available_value < target_value)
            # Selected value is out of range, go back and try other branch
            or (current_value > target_value + cost_of_change)
            # Don't select things which we know will be more wasteful if the waste is increasing
            or (current_waste > best_waste and utxo_pool[0].fee - utxo_pool[0].long_term_fee > 0)
        ):
            should_backtrack = True
        # Selected value is within range
        elif current_value >= target_value:
            # This^^ is the excess value which is added to the waste for the below comparison
            # Adding another UTXO after this check could bring the waste down if the long term fee is higher than the current fee.
            # However we are not going to explore that because this optimization for the waste is only done when we have hit our target
            # value. Adding any more UTXOs will be just burning the UTXO; it will go entirely to fees. Thus we aren't going to
            # explore any more UTXOs to avoid burning money like that.

            current_waste += (current_value - target_value)
            if current_waste <= best_waste:
                best_selection = current_selection.copy()
                best_waste = current_waste
                if (best_waste == 0):
                    return CoinSelection.from_utxo_pool(best_selection, utxo_pool)
            # Remove the excess value as we will be selecting different coins now
            current_waste -= (current_value - target_value)
            should_backtrack = True
        # Backtracking, moving backwards
        if should_backtrack:
            # Walk backwards to find the last included UTXO
            # that still needs to have its omission branch traversed
            while len(current_selection) > 0 and current_selection[-1] == False:
                current_selection.pop()
                current_available_value += utxo_pool[len(
                    current_selection)].effective_value
            if len(current_selection) == 0:
                # We have walked back to the first utxo and no branch is untraversed.
                # All solutions searched
                break

            # Output was included on previous iterations, try excluding now
            current_selection[-1] = False
            utxo: OutputGroup = utxo_pool[len(current_selection)-1]
            current_value -= utxo.effective_value
            current_waste -= (utxo.fee - utxo.long_term_fee)
        # Moving forwards, continuing down this branch
        else:
            utxo = utxo_pool[len(current_selection)]
            current_available_value -= utxo.effective_value
            # Avoid searching a branch if the previous UTXO has the same value and
            # same waste and was excluded. Since the ratio of fee to long term fee
            # is the same, we only need to check if one of those values match in
            # order to know that the waste is the same
            if (
                len(current_selection) > 0
                and current_selection[-1] == False
                and utxo.effective_value == utxo_pool[len(current_selection) - 1].effective_value
                and utxo.fee == utxo_pool[len(current_selection) - 1].fee
            ):
                current_selection.append(False)
            else:
                # Inclusion branch first (Largest First Exploration)
                current_selection.append(True)
                current_value += utxo.effective_value
                current_waste += (utxo.fee - utxo.long_term_fee)

    # Check for solution
    if len(best_selection) == 0:
        return CoinSelection.algorithm_failure()

    # Set output set
    return CoinSelection.from_utxo_pool(best_selection, utxo_pool)



from typing import List

from bitcoin_coin_selection.selection_algorithms.branch_and_bound import select_coins_branch_and_bound
from bitcoin_coin_selection.selection_algorithms.knapsack_solver import select_coins_knapsack_solver
from bitcoin_coin_selection.selection_algorithms.single_random_draw import select_coins_single_random_draw
from bitcoin_coin_selection.selection_types.change_constants import MAX_MONEY
from bitcoin_coin_selection.selection_types.output_group import OutputGroup
from bitcoin_coin_selection.selection_types.coin_selection import (
    CoinSelection
)


def select_coins(utxo_pool: List[OutputGroup],
                 target_value: int,
                 short_term_fee_per_byte: int,
                 long_term_fee_per_byte: int,
                 change_output_size_in_bytes: int,
                 change_spend_size_in_bytes: int,
                 not_input_size_in_bytes: int
                 ) -> CoinSelection:

    # Validate target value isn't something silly
    if target_value == 0 or target_value > MAX_MONEY:
        return CoinSelection.invalid_spend()

    # Check for insufficient funds
    total_value = int(sum(
        [output_group.value for output_group in utxo_pool]))
    if total_value < target_value:
        return CoinSelection.insufficient_funds()

    # Calculate fees spending any given utxo would incur
    for outut_group in utxo_pool:
        outut_group.set_fee(short_term_fee_per_byte, long_term_fee_per_byte)
    # Calculate fee for the "fixed" part of a transaction
    fixed_fee = short_term_fee_per_byte * not_input_size_in_bytes
    target_after_fixed_fees = target_value + fixed_fee

    # Check for insufficient funds after fees
    total_effective_value = int(sum(
        [output_group.effective_value for output_group in utxo_pool]))
    if total_effective_value < target_after_fixed_fees:
        return CoinSelection.insufficient_funds_after_fees()

    # Calculate cost of change (input to branch and bound)
    cost_of_creating_change = short_term_fee_per_byte * change_output_size_in_bytes
    cost_of_spending_change = long_term_fee_per_byte * change_spend_size_in_bytes
    cost_of_change = cost_of_creating_change + cost_of_spending_change

    # Return branch and bound selection (more optimized) if possible
    bnb_selection = select_coins_branch_and_bound(
        utxo_pool, target_after_fixed_fees, cost_of_change)
    if bnb_selection.outcome == CoinSelection.Outcome.SUCCESS:
        return bnb_selection
    # Otherwise return knapsack_selection (less optimized) if possible
    else:
        knapsack_selection = select_coins_knapsack_solver(
            utxo_pool, target_after_fixed_fees
        )
        if knapsack_selection.outcome == CoinSelection.Outcome.SUCCESS:
            return knapsack_selection
        else:
            # If all else fails, return single random draw selection (not optomized) as a fallback
            return select_coins_single_random_draw(utxo_pool, target_after_fixed_fees)



import random
from typing import List

from bitcoin_coin_selection.selection_types.coin_selection import CoinSelection

from bitcoin_coin_selection.selection_types.output_group import OutputGroup


def select_coins_single_random_draw(
        utxo_pool: List[OutputGroup], target_value: int) -> CoinSelection:

    random.shuffle(utxo_pool)
    selected_output_groups = []
    selected_value = 0
    for output_group in utxo_pool:
        selected_value += output_group.effective_value
        selected_output_groups.append(output_group)
        if selected_value >= target_value:
            return CoinSelection(selected_output_groups)

    return CoinSelection.algorithm_failure()



from typing import List
import pytest

from bitcoin_coin_selection.selection_algorithms.single_random_draw import select_coins_single_random_draw
from bitcoin_coin_selection.tests.fixtures import generate_utxo_pool
from bitcoin_coin_selection.selection_types.change_constants import CENT
from bitcoin_coin_selection.selection_types.coin_selection import CoinSelection


RUN_TESTS = 100


def test_single_random_draw_failure_1(generate_utxo_pool):
    utxo_pool = generate_utxo_pool([i * CENT for i in range(100)])
    for i in range(RUN_TESTS):
        selection = select_coins_single_random_draw(
            utxo_pool, 100000 * CENT)
        assert selection.outcome == CoinSelection.Outcome.ALGORITHM_FAILURE


def test_single_random_draw_failure_2(generate_utxo_pool):
    utxo_pool = generate_utxo_pool([])
    selection = select_coins_single_random_draw(
        utxo_pool, 1)
    assert selection.outcome == CoinSelection.Outcome.ALGORITHM_FAILURE


def test_single_random_draw_success_1(generate_utxo_pool):
    utxo_pool = generate_utxo_pool([i * CENT for i in range(100)])
    for i in range(RUN_TESTS):
        selection = select_coins_single_random_draw(
            utxo_pool, 150 * CENT)

        assert selection.outcome == CoinSelection.Outcome.SUCCESS
        assert selection.effective_value >= 150 * CENT


def test_single_random_draw_success_2(generate_utxo_pool):
    utxo_pool = generate_utxo_pool([1 * CENT])
    selection = select_coins_single_random_draw(
        utxo_pool, 1 * CENT)

    assert selection.outcome == CoinSelection.Outcome.SUCCESS
    assert selection.effective_value == 1 * CENT



from enum import Enum
from typing import List


from bitcoin_coin_selection.selection_types.input_coin import InputCoin
from bitcoin_coin_selection.selection_types.output_group import OutputGroup


class CoinSelection():

    class Outcome(Enum):
        SUCCESS = 0
        INSUFFICIENT_FUNDS = 1
        INSUFFICIENT_FUNDS_AFTER_FEES = 2
        ALGORITHM_FAILURE = 3
        INVALID_SPEND = 4

    outcome: Outcome
    target: int
    outputs: List[InputCoin]
    value: int
    effective_value: int
    fee: int

    def __init__(self, selected_output_groups: List[OutputGroup] = None, outcome=Outcome.SUCCESS):
        self.outputs = []
        self.value = 0
        self.effective_value = 0
        self.fee = 0
        self.outcome = outcome
        if selected_output_groups:
            for output_group in selected_output_groups:
                for output in output_group.outputs:
                    self.insert(output_group)

    @classmethod
    def insufficient_funds(cls):
        return cls(None, cls.Outcome.INSUFFICIENT_FUNDS)

    @classmethod
    def insufficient_funds_after_fees(cls):
        return cls(None, cls.Outcome.INSUFFICIENT_FUNDS_AFTER_FEES)

    @classmethod
    def algorithm_failure(cls):
        return cls(None, cls.Outcome.ALGORITHM_FAILURE)

    @classmethod
    def invalid_spend(cls):
        return cls(None, cls.Outcome.INVALID_SPEND)

    @classmethod
    def from_utxo_pool(cls, best_selection: List[bool], utxo_pool: List[OutputGroup]):
        selected_groups = []
        for i, was_selected in enumerate(best_selection):
            if was_selected:
                selected_group = utxo_pool[i]
                selected_groups.append(selected_group)

        return cls(selected_groups)

    def insert(self, output: InputCoin):
        self.outputs.append(output)
        self.value += output.value
        self.effective_value += output.effective_value
        self.fee += output.fee
